/*
======================================================
====      Java Classes For Genetic Algorithms     ====
==== (c) Marek Obitko, 1998-1999, obitko@email.cz ====
======================================================

This file is a part of Java library for genetic algorithms with 
visualization. You may use this library or its part free of charge
provided you acknowledge the author (Marek Obitko), and the source 
(Learning Environment for Genetic Algorithms, one of the URLs 
below) in your work where you use it (contact the author, if you 
are in doubt). Moreover I would appreciate if you inform me 
(via email) about your work where you used it 
(this is not compulsory).

THIS LIBRARY IS PROVIDED "AS IS", WITHOUT ANY WARRANTY.

This package was designed for learning environment for GAs, see
http://cs.felk.cvut.cz/~xobitko/ga  or

Author is studying Czech Technical University in Prague, 
Faculty of Electrical Engineering, Department of Computer Engineering.
Major part of this library was written at University of Applied Sciency
in Dresden, Germany.
*/

/** Stores information about creature movement. Used for GA developing movement. 
  * @see WalkGA
  * <p>
  * @author Marek Obitko, <A HREF="mailto:obitko@email.cz">obitko@email.cz</A>
  * @version 1.0
  */
public class CreatureChromosome extends Chromosome
{

  /** No of joints. */
  int joints;
  
  /** No of positions. */
  int positions;
  
  /** Chromosome information - movement variations */
  double v[][];
  
  Creature mc;
  
  public CreatureChromosome(int aJoints, int aPositions, Creature creature)
  {
    joints=aJoints;
    positions=aPositions;
    mc=creature;
    v=new double[joints][positions];
  }
  
  public void countFitness()
  {
    fitness=mc.evaluate(this);
  }
  
  public Chromosome[] crossover(Chromosome other, double probability)
  {
    CreatureChromosome[] result=new CreatureChromosome[2];

    if(probability<(Math.random()*100))
       {
         lastCrossoverPoint=-1;
         result[0]=(CreatureChromosome)((CreatureChromosome)this).copy();
         result[1]=(CreatureChromosome)((CreatureChromosome)other).copy();
       }  
     else
      {
        result[0]=new CreatureChromosome(joints,positions,mc);
        result[1]=new CreatureChromosome(joints,positions,mc);
        
        for(int i=0; i<joints; i++)
          for(int j=0; j<positions; j++)    
            if(Math.random()<probability) 
              v[i][j]+=(Math.random()*2-1)/WalkParams.MUTATION_VARIATION;
        
        boolean order=true;
         
        for(int i=0; i<joints; i++)
          for(int j=0; j<positions; j++)    
            {
              if(Math.random()<WalkParams.CROSSOVER_SWITCH) order=!order;
              if(order)
                {
                  result[0].v[i][j]=this.v[i][j];
                  result[1].v[i][j]=((CreatureChromosome)other).v[i][j];
                } else {
                  result[1].v[i][j]=this.v[i][j];
                  result[0].v[i][j]=((CreatureChromosome)other).v[i][j];
                }
                
            }
      }
    return result;
  }
  
  public void mute(double probability)
  {
    for(int i=0; i<joints; i++)
      for(int j=0; j<positions; j++)    
        if(Math.random()*100<probability) 
          v[i][j]+=((Math.random()*2-1)*WalkParams.VAR_SIZE)/WalkParams.MUTATION_VARIATION;
  }
  
  public void randomize()
  {
    for(int i=0; i<joints; i++)    
      for(int j=0; j<positions; j++)    
        v[i][j]=(Math.random()*2-1)*WalkParams.VAR_SIZE;
  }
  
  public Chromosome copy()
  {
    CreatureChromosome result=new CreatureChromosome(joints, positions, mc);
    for(int i=0; i<joints; i++)    
      for(int j=0; j<positions; j++)    
        result.v[i][j]=v[i][j];
    return result;
  }

}