/*
======================================================
====      Java Classes For Genetic Algorithms     ====
==== (c) Marek Obitko, 1998-1999, obitko@email.cz ====
======================================================

This file is a part of Java library for genetic algorithms with 
visualization. You may use this library or its part free of charge
provided you acknowledge the author (Marek Obitko), and the source 
(Learning Environment for Genetic Algorithms, one of the URLs 
below) in your work where you use it (contact the author, if you 
are in doubt). Moreover I would appreciate if you inform me 
(via email) about your work where you used it 
(this is not compulsory).

THIS LIBRARY IS PROVIDED "AS IS", WITHOUT ANY WARRANTY.

This package was designed for learning environment for GAs, see
http://cs.felk.cvut.cz/~xobitko/ga  or

Author is studying Czech Technical University in Prague, 
Faculty of Electrical Engineering, Department of Computer Engineering.
Major part of this library was written at University of Applied Sciency
in Dresden, Germany.
*/

import java.awt.Graphics;
import java.awt.Canvas;
import java.awt.Dimension; 
import java.awt.Color;
import java.awt.Image;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseListener;

/** Component for drawing 3D graph and solutions on it.
  * <p> It uses double buffering to prevent blinking.
  * @see Function3D
  * <p>
  * @author Marek Obitko, <A HREF="mailto:obitko@email.cz">obitko@email.cz</A>
  * @version 1.0
  */
public class Graph3DCanvas extends Canvas 
                           implements MouseMotionListener,  MouseListener
{

  /** Size of component */
  protected Dimension mySize;
  
  /** Population being displayed on the graph. */
  protected Population population=null;
  
  /** Graphics of component */
  protected Graphics gr;
  
  /** Graphics of offline buffer */
  protected Graphics offgr;
  
  /** Image of offline buffer */
  protected Image offscreen;
  
  Graph3D g3d;
  
  String theBest=null;
    
  /** Constructor which makes standart size graph */
  public Graph3DCanvas(Graph3D aGraph3D)
  {
    this(350,350, aGraph3D);
  }
  
  /** Constructor where you can specify size of component */
  public Graph3DCanvas(int xsize, int ysize, Graph3D aGraph3D)
  {
    super();
    mySize=new Dimension(xsize,ysize);
    g3d=aGraph3D;
    g3d.cSize=xsize;
    g3d.recalculate();
    addMouseMotionListener(this);
    addMouseListener(this);
  }

  /** Draws graph on graphics.
    * @param g graphics which the graph is drawn on
    */
  protected void drawGraph(Graphics g)
  {
    if(!g3d.valid) return;
    
    drawBox(g);
    
    int lx=0, ly=0, nx, ny;
    int i,j;
    for(i=0; i<g3d.steps; i+=g3d.bigStep)
      for(j=0; j<g3d.steps; j++)
        {
          nx = g3d.x[i][j];
          ny = g3d.y[i][j];
          if(j!=0) g.drawLine(lx,ly,nx,ny);
          lx=nx;
          ly=ny;
        }
    i=g3d.steps-1;
      for(j=0; j<g3d.steps; j++)
        {
          nx = g3d.x[i][j];
          ny = g3d.y[i][j];
          if(j!=0) g.drawLine(lx,ly,nx,ny);
          lx=nx;
          ly=ny;
        }
    for(j=0; j<g3d.steps; j+=g3d.bigStep)
      for(i=0; i<g3d.steps; i++)
        {
          nx = g3d.x[i][j];
          ny = g3d.y[i][j];
          if(i!=0) g.drawLine(lx,ly,nx,ny);
          lx=nx;
          ly=ny;
        }
    j=g3d.steps-1;
      for(i=0; i<g3d.steps; i++)
        {
          nx = g3d.x[i][j];
          ny = g3d.y[i][j];
          if(i!=0) g.drawLine(lx,ly,nx,ny);
          lx=nx;
          ly=ny;
        }
  }
  
  /** Draws box around graph. */
  protected void drawBox(Graphics gr)
  {
    long cor[]=new long[2];
    int a[]=new int[2];
    int b[]=new int[2];
    int c[]=new int[2];
    int d[]=new int[2];
    int e[]=new int[2];
    int f[]=new int[2];
    int g[]=new int[2];
    int h[]=new int[2];
    gr.setColor(Color.cyan);
        
    cor[0]=0;
    cor[1]=0;
    g3d.xyBottomCoords(cor,a);
    g3d.xyTopCoords(cor,d);
    cor[1]=4294967296L;
    g3d.xyBottomCoords(cor,e);
    g3d.xyTopCoords(cor,h);
    cor[0]=4294967296L;
    g3d.xyBottomCoords(cor,f);
    g3d.xyTopCoords(cor,g);
    cor[1]=0;
    g3d.xyBottomCoords(cor,b);
    g3d.xyTopCoords(cor,c);
    
    gr.drawLine(a[0],a[1],d[0],d[1]);
    gr.drawLine(a[0],a[1],e[0],e[1]);
    gr.drawLine(a[0],a[1],b[0],b[1]);
    gr.drawLine(g[0],g[1],f[0],f[1]);
    gr.drawLine(g[0],g[1],h[0],h[1]);
    gr.drawLine(g[0],g[1],c[0],c[1]);
    gr.drawLine(d[0],d[1],c[0],c[1]);
    gr.drawLine(d[0],d[1],h[0],h[1]);
    gr.drawLine(b[0],b[1],c[0],c[1]);
    gr.drawLine(b[0],b[1],f[0],f[1]);
    gr.drawLine(e[0],e[1],f[0],f[1]);
    gr.drawLine(e[0],e[1],h[0],h[1]);
    
    gr.setColor(Color.black);
  }

  /** Draws population on graphics.
    * @param g graphics which the solutions are drawn on
    */
  protected void drawPopulation(Graphics g)
  {
    if(population==null) return;
    
    long coords[]=new long[2];
    int xyb[]=new int[2];
    int xyt[]=new int[2];
    int xyz[]=new int[2];
    
    g.setColor(Color.blue);
    for(int i=1; i<population.size; i++)
      {
        coords[0]=((Number3DChromosome)(population.pop[i])).xvalue();
        coords[1]=((Number3DChromosome)(population.pop[i])).yvalue();
        g3d.xyBottomCoords(coords,xyb);
        g3d.xyTopCoords(coords,xyt);
        g3d.xyZCoords(coords,xyz);
        g.drawLine(xyb[0],xyb[1],xyt[0],xyt[1]);
        g.fillRect(xyz[0]-2,xyz[1]-2,5,5);
      }
    g.setColor(Color.red);  // the best is red
    coords[0]=((Number3DChromosome)(population.pop[0])).xvalue();
    coords[1]=((Number3DChromosome)(population.pop[0])).yvalue();

    StringBuffer sbf = new StringBuffer(""+g3d.xValue(coords[0]));
    sbf.setLength(11);
    String s="Best [X,Y,Z]=["+sbf+"; ";
    sbf = new StringBuffer(""+g3d.yValue(coords[1]));
    sbf.setLength(11);
    s=s+sbf+"; ";    
    sbf = new StringBuffer(""+g3d.zValue(coords[0], coords[1]));
    sbf.setLength(11);
    s=s+sbf+"]";    

    g3d.xyBottomCoords(coords,xyb);
    g3d.xyTopCoords(coords,xyt);
    g3d.xyZCoords(coords,xyz);
    theBest=s;
    g.drawLine(xyb[0],xyb[1],xyt[0],xyt[1]);
    g.fillRect(xyz[0]-2,xyz[1]-2,5,5);
    g.setColor(Color.black);
  }
  
  /** Initializes double buffering.
    * This should be called before using method paint.
    * @see #paint
    */
  public void initPaint()
  {
    gr = this.getGraphics();
    offscreen = this.createImage(mySize.width, mySize.height);
    offgr = offscreen.getGraphics();
  }
  
  /** This method acts like repaint(), but it prevents blinking by buffering
    * an image.
    */
  public synchronized void paint()
  {
    if(offgr==null) initPaint();
    offgr.setColor(getBackground());
    offgr.fillRect(0, 0, mySize.width, mySize.height);
    offgr.setColor(getForeground());
    paint(offgr);
    gr.drawImage(offscreen, 0, 0, this);
  }
  
  /** Paints graph and solutions on it (if there are any). */
  public synchronized void paint(Graphics g) 
  {
    g.setColor(Color.black);
    drawGraph(g);  
    drawPopulation(g); 
    if(Function3D.wasError)  // box around - red if error, otherwise black
      {
        g.setColor(Color.red);
        g.drawRect(0, 0, mySize.width - 1, mySize.height - 1);
        g.drawRect(1, 1, mySize.width - 3, mySize.height - 3);
        g.drawString("Please check inputs.",10,20);
        //g.drawString("I don't understand what you mean.",10,40);
        g.setColor(Color.black);  
      } else {
        g.setColor(Color.black);
        g.drawRect(0, 0, mySize.width - 1, mySize.height - 1);
      }  
  }
  
  /** Recalculates and redraws graph. */
  public synchronized void resetAll()
  {
    g3d.recalculate();
    paint();
  }
  
  int mouseX, mouseY;
  protected final static double SHIFT_FACTOR=Math.PI/180/2;
  
  protected boolean mouseDown=false;
  
  public synchronized void mouseDragged(MouseEvent e) 
  {
    if(!mouseDown) return;
    
    double thetaA=(e.getX()-mouseX)*SHIFT_FACTOR;
    double phiA=(e.getY()-mouseY)*SHIFT_FACTOR;
    g3d.theta-=thetaA;
    g3d.phi+=phiA;  
      
    mouseX=e.getX(); 
    mouseY=e.getY();
    
    g3d.recalculate();
    paint();
  }

  public void mouseMoved(MouseEvent e) 
  {
  }

  public void mouseClicked (MouseEvent e)  
  {
  }
 
  public void mouseEntered (MouseEvent e)  
  {
  }
 
  public void mouseExited  (MouseEvent e)  
  {
  }
 
  public void mousePressed (MouseEvent e)  
  {
    mouseX=e.getX(); 
    mouseY=e.getY();
    mouseDown=true;
  }
  
  public void mouseReleased(MouseEvent e)
  {
    mouseDown=false;
  }
 
  
  /** Sets population for drawing.
    * @param aPopulation population, which will be shown on the graph; use null for
    *        drawing no population on graph
    */ 
  public void setPopulation(Population aPopulation)
  {
    population=aPopulation;
  }
  
  public Dimension getPreferredSize()
  {
    return mySize;
  }

  public Dimension getMaximumSize()
  {
    return mySize;
  }

  public Dimension getMinimumSize()
  {
    return mySize;
  }

}