/*
======================================================
====      Java Classes For Genetic Algorithms     ====
==== (c) Marek Obitko, 1998-1999, obitko@email.cz ====
======================================================

This file is a part of Java library for genetic algorithms with 
visualization. You may use this library or its part free of charge
provided you acknowledge the author (Marek Obitko), and the source 
(Learning Environment for Genetic Algorithms, one of the URLs 
below) in your work where you use it (contact the author, if you 
are in doubt). Moreover I would appreciate if you inform me 
(via email) about your work where you used it 
(this is not compulsory).

THIS LIBRARY IS PROVIDED "AS IS", WITHOUT ANY WARRANTY.

This package was designed for learning environment for GAs, see
http://cs.felk.cvut.cz/~xobitko/ga  or

Author is studying Czech Technical University in Prague, 
Faculty of Electrical Engineering, Department of Computer Engineering.
Major part of this library was written at University of Applied Sciency
in Dresden, Germany.
*/

import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;


/** Shows basic genetic algorithm on solving TSP (Travelling salesman problem).
  * <p>
  * Allows to see WHOLE population and how it is developing.
  * <p>
  * @author Marek Obitko, <A HREF="mailto:obitko@email.cz">obitko@email.cz</A>
  * @version 1.0
  */
public class TSPGAAll extends Applet implements ActionListener, ItemListener, Runnable
{
  /** Pause between generations in run of algorithm */
  final private int STEP_PAUSE=0;
  
  /** Size of population (number of chromosomes), should be even number */
  final private int POP_SIZE=16;
                                 
  /** Thread for running of genetic algorithm in background */
  Thread algThread=null;
                                   
  /** Population of solutions in shown genetic algorithm */
  Population population;

  TSPInstance points=new TSPInstance(100);
  
  protected boolean gaInitialized;
  
  /** Component for drawing graph and solutions on it */
  TSPPathCanvas Path[]; 
  /** Component for drawing graph and solutions on it */
  TSPPathCanvas BestPath; 
    
  /** Button Start */
  Button Start=new Button(" Start ");

  /** Button Step */
  Button Step=new Button(" Step ");

  /** Button Stop */
  Button Stop=new Button(" Stop ");

  /** Button Reset */
  Button Reset=new Button(" Reset ");
  
  Button View=new Button("Change View");
  
  TextField popNo = new TextField("0",5);
  
  Checkbox showNumbers = new Checkbox("Numbers");

  Choice mutationC=new Choice();
  Choice crossoverC=new Choice();

  public int crossoverType=Chromosome.BASIC_CROSSOVER;
  
  public int mutationType=Chromosome.BASIC_MUTATION;

  CardLayout viewLayout = new CardLayout();
  Panel pv=new Panel(viewLayout);
  

  /** Initialization of applet and genetic algorithm */
  public void init()
  {
    setBGColor();
    gaInitialized=false;
    population=null;
    Path=new TSPPathCanvas[POP_SIZE];
        
    GridBagLayout gbl=new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.insets=new Insets(0,0,0,0);
    Panel p16=new Panel(gbl);

    for(int i=0; i<4; i++)
    for(int j=0; j<4; j++)
      {
        Path[i*4+j]=new TSPPathCanvas(120,120,points);
        gbc.fill=GridBagConstraints.CENTER;
        gbc.gridx=j;
        gbc.gridy=i;
        gbl.setConstraints(Path[i*4+j], gbc);
        p16.add(Path[i*4+j]);
        //Path[i*4+j].randomize(5);
        Path[i*4+j].drawPath=true;
        Path[i*4+j].parentApplet=this;
        //        Path[i*4+j].setPathLength(0);
      }
    Path[0].randomize(5);  
    Path[0].drawPath=true;
    pv.add("one",p16);
    
    BestPath=new TSPPathCanvas(480,480,points);
    Panel pb=new Panel(new BorderLayout());
    pb.add("Center",BestPath);
    pv.add("two",pb);
    BestPath.parentApplet=this;
    BestPath.canAdd=true;
    BestPath.drawPath=true;
    BestPath.BORDER=4;
    // BestPath.setPathLength(0);
    //BestPath.randomize(5);
    add(pv);
    
    Panel pc=new Panel();
    pc.add(Start);
    Start.addActionListener(this);
    pc.add(Step);
    Step.addActionListener(this);
    pc.add(Stop);  
    Stop.addActionListener(this);
    pc.add(Reset);
    Reset.addActionListener(this);
    pc.add(new Label(" "));
    pc.add(View);
    View.addActionListener(this);
    pc.add(new Label("  Generation", Label.RIGHT));
    pc.add(popNo);
    add(pc);
    
    Panel cp=new Panel();
    cp.add(showNumbers);
    showNumbers.setState(true);
    showNumbers.addItemListener(this);
    crossoverC.add("One point crossover");
    crossoverC.add("Two point crossover");
    crossoverC.add("No crossover");
    cp.add(crossoverC);
    crossoverC.addItemListener(this);
    mutationC.add("Normal random mutation");
    mutationC.add("Random only improving mutation");
    mutationC.add("Systematic only improving mutation");
    mutationC.add("Random improving mutation");
    mutationC.add("Systematic improving mutation");
    mutationC.add("No mutation");
    cp.add(mutationC);
    mutationC.addItemListener(this);
    add(cp);
    //actionChangeView();
  }

  /** Sets the background color, which is specified in Applet parameter BGCOLOR.
    * For using, just add to applet tag in HTML something like
    * "<"PARAM name=BGCOLOR value="00ff40">".
    */
  protected void setBGColor()
  {
    try 
      {
        Color bg = new Color(Integer.parseInt(getParameter("BGCOLOR"), 16));
        setBackground(bg);
      } catch (NumberFormatException e) {}
        catch (NullPointerException e) {}
  }
    
  /** Handles actions from checkboxes and item selectors */
  public void itemStateChanged(ItemEvent e) 
  {
    ItemSelectable is = e.getItemSelectable();
    if(is==showNumbers)
      for(int i=0; i<POP_SIZE; i++)
        {
          Path[i].DRAW_NUMBERS=showNumbers.getState();
          Path[i].paint();
          BestPath.DRAW_NUMBERS=showNumbers.getState();
          BestPath.paint();
        }
    if(is==mutationC)
      { 
        mutationType=mutationC.getSelectedIndex();
        if(population!=null) population.mutationType=mutationType;
      }  
    if(is==crossoverC) 
      {  
        crossoverType=crossoverC.getSelectedIndex();
        if(population!=null) population.crossoverType=crossoverType;
      }  
  }
  
  /** Handles actions from buttons */
  public void actionPerformed(ActionEvent e)
  {
    String command=e.getActionCommand();
    if(command.equals(Start.getLabel())) actionStart();
    if(command.equals(Stop.getLabel())) actionStop();
    if(command.equals(Reset.getLabel())) actionReset();
    if(command.equals(Step.getLabel())) actionStep();
    if(command.equals(View.getLabel())) actionChangeView();
    if(command.equals("path")) actionInit(); // change of TSP Instance
  }

  /** Button Start starts genetic algorithm (as a thread) */
  protected void actionStart()
  {
    if(algThread!=null) return;
    algThread = new Thread(this);
    algThread.start();
  }

  protected boolean panel16=true;
  
  /** Changes view between the best chromosome and the whole population. */
  protected synchronized void actionChangeView()
  {
    if(panel16)
      {
        //viewLayout.show(pv,"two");
        viewLayout.last(pv);
        points.changeSize(120,480);
        //BestPath.enableMouseEvents();
      }
     else
      {   
        //viewLayout.show(pv,"one");
        viewLayout.first(pv);
        points.changeSize(480,120);
        //for(int i=0; i<POP_SIZE; i++) 
        //     Path[i].enableMouseEvents();
      }    
    //viewLayout.invalidateLayout(pv);
    panel16=!panel16;
  }
  
  /** Button Reset resets genetic algorithm */
  protected synchronized void actionReset()
  {
    if(!gaInitialized) actionInit();
    population.randomize();
    for(int i=0; i<POP_SIZE; i++){
      Path[i].setPath(((TSPChromosome)(population.pop[i])).perm);
      //      Path[i].setPathLength(Integer.MAX_VALUE-((TSPChromosome)(population.pop[i])).fitness);
    }
    BestPath.setPath(((TSPChromosome)(population.pop[0])).perm);
    // BestPath.setPathLength(Integer.MAX_VALUE-((TSPChromosome)(population.pop[0])).fitness);
  }
  
  /** Button Stop stops genetic algorithm (the thread) */
  protected void actionStop()
  {
    //Path.drawPath=false;
    if(algThread!=null)
      {
        algThread.stop();
        algThread=null;
      }
    for(int i=0; i<POP_SIZE; i++)
      {
        Path[i].paint();
        Path[i].canAdd=true;
      }
    BestPath.paint();
    BestPath.canAdd=true;
  }
  
  /** Button Step performs one step (on generation) of genetic algorithm */
  protected synchronized void actionStep()
  {
    if(!gaInitialized) actionInit();
    //if(!points.dmComputed) actionInit();
    population.newGeneration();
    for(int i=0; i<POP_SIZE; i++)
      {
        Path[i].setPath(((TSPChromosome)(population.pop[i])).perm);
        //Path[i].paint();
      }  
    BestPath.setPath(((TSPChromosome)(population.pop[0])).perm);
    //BestPath.paint();
  }
  
  /** Initializes (and before it creates) whole population. */
  protected void actionInit()
  {
    //System.out.println("INIT");
    for(int i=0; i<POP_SIZE; i++)
      Path[i].changed(); // change will be handled now, so clear the semaphore
    BestPath.changed();  
    //Path.canAdd=false;
    points.computeDistanceMatrix();
    TSPChromosome sample=new TSPChromosome(points.size, points);
    sample.randomize();
    sample.countFitness();
    population=new Population(sample, POP_SIZE);
    population.crossoverType=crossoverType;
    population.mutationType=mutationType;
    population.popNo=popNo;
    population.randomize();
    for(int i=0; i<POP_SIZE; i++)
      {
        Path[i].setPath(((TSPChromosome)(population.pop[i])).perm);
        Path[i].drawPath=true;
        Path[i].paint();
      }
    BestPath.setPath(((TSPChromosome)(population.pop[0])).perm);
    BestPath.drawPath=true;
    BestPath.paint();
    //viewLayout.invalidateLayout(pv);
    gaInitialized=true;
  }
  
  /** Defines run of genetic algorithm thread */
  public void run()
  {
    //Path.setPopulation(population);
    //Path.paint();
    //if(!gaInitialized) actionInit();
    //try { Thread.sleep(STEP_PAUSE);
    //    } catch (InterruptedException e) {}
    while(true)
      {
         actionStep();   
         try { Thread.sleep(STEP_PAUSE);
             } catch (InterruptedException e) {}
      }
  }

  /** Finishing of the applet and destroying thread. */
  public void destroy()
  {
    if(algThread!=null)
      {
        algThread.stop();
        algThread=null;
      }
    super.destroy();
  }

  /** Finishing of the applet and destroying thread.
    * @exception Throwable successor can throw an exception
    */
  protected void finalize() throws Throwable
  {
    if(algThread!=null)
      {
        algThread.stop();
        algThread=null;
      }
    super.finalize();
  }

  /** Returns information about parameters of applet. */
  public String[][] getParameterInfo() 
  {
    String[][] info = 
     {
      {"BGCOLOR", "color (6 hexa digits)", "Background color, for example \"00ff40\"."},
     };
        return info;
  }

  /** Returns information about applet. */
  public String getAppletInfo() 
  {
    return "Title: Genetic algorithm for TSP \n"+
           "Author: Marek Obitko, obitko@email.cz \n"+
           "Applet shows GA solving travelling salesman problem. It allows to choose operators and"+
           "enables to show whole population.";
  }


  /** For running of applet without browser or appletviewer (it is then possible
    * to run this applet as a class directly).
    * @see GAApplication
    */
  public static void main(String[] args)
  {
    Applet applet = new BasicGA();
    Frame frame = new GAApplication("Genetic Algorithm for TSP", applet, 720, 250);
  }
}
