/*
======================================================
====      Java Classes For Genetic Algorithms     ====
==== (c) Marek Obitko, 1998-1999, obitko@email.cz ====
======================================================

This file is a part of Java library for genetic algorithms with 
visualization. You may use this library or its part free of charge
provided you acknowledge the author (Marek Obitko), and the source 
(Learning Environment for Genetic Algorithms, one of the URLs 
below) in your work where you use it (contact the author, if you 
are in doubt). Moreover I would appreciate if you inform me 
(via email) about your work where you used it 
(this is not compulsory).

THIS LIBRARY IS PROVIDED "AS IS", WITHOUT ANY WARRANTY.

This package was designed for learning environment for GAs, see
http://cs.felk.cvut.cz/~xobitko/ga  or

Author is studying Czech Technical University in Prague, 
Faculty of Electrical Engineering, Department of Computer Engineering.
Major part of this library was written at University of Applied Sciency
in Dresden, Germany.
*/

import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;


/** Shows genetic algorithm developing movement of a creature.
  * <p>
  * IT IS NOT FINISHED YET.
  * @see Creature
  * @see Joint
  * @see CreatureChromosome
  * @see CreatureInfo
  * @see WalkParams
  * @see WalkCanvas
  * <p>
  * @author Marek Obitko, <A HREF="mailto:obitko@email.cz">obitko@email.cz</A>
  * @version 1.0
  */
public class WalkGA extends Applet implements ActionListener, Runnable
{
  /** Pause between generations in run of algorithm */
  final private int STEP_PAUSE=0;
  
  
  /** Thread for running of genetic algorithm in background */
  Thread algThread=null;
                                   
  /** Population of solutions in shown genetic algorithm */
  Population population;

  /** Component for drawing graph and solutions on it */
  WalkCanvas WCanvas=new WalkCanvas();
  
  /** Button Start */
  Button Start=new Button(" Start ");

  /** Button Step */
  Button Step=new Button(" Step ");

  /** Button Stop */
  Button Stop=new Button(" Stop ");

  /** Button Reset */
  Button Reset=new Button(" Reset ");

  Creature myCreature;
  CreatureInfo crs=new CreatureInfo();
  
  /** Initialization of applet and genetic algorithm */
  public void init() 
  {
    setBGColor();
    myCreature=crs.readCreature();
    myCreature.createPopulation(WalkParams.POSITIONS);
    
    population=myCreature.population;
    //population.randomize();
    add(WCanvas);
    add(Start); 
    Start.addActionListener(this);
    add(Step);
    Step.addActionListener(this);
    add(Stop);
    Stop.addActionListener(this);
    add(Reset);
    Reset.addActionListener(this);
    WCanvas.initPaint(); 
  }
    
  
  /** Sets the background color, which is specified in Applete parameter BGCOLOR.
    * For using, just add to applet tag in HTML something like
    * "<"PARAM name=BGCOLOR value="00ff40">".
    */
  protected void setBGColor()
  {
    try 
      {
        Color bg = new Color(Integer.parseInt(getParameter("BGCOLOR"), 16));
        setBackground(bg);
      } catch (NumberFormatException e) {}
        catch (NullPointerException e) {}
  }
  
  /** Handles actions from buttons */
  public void actionPerformed(ActionEvent e)
  {
    String command=e.getActionCommand();
    if(command.equals(Start.getLabel())) actionStart();
    if(command.equals(Stop.getLabel())) actionStop();
    if(command.equals(Reset.getLabel())) actionReset();
    if(command.equals(Step.getLabel())) actionStep();
  }

  /** Button Start starts genetic algorithm (as a thread) */
  protected void actionStart()
  {
    if(algThread!=null) return;
    algThread = new Thread(this);
    algThread.start();
  }
  
  /** Button Reset resets genetic algorithm */
  protected void actionReset()
  {
    population.randomize();
    WCanvas.paint();
  }
  
  /** Button Stop stops genetic algorithm (the thread) */
  protected void actionStop()
  {
    if(algThread!=null)
      {
        algThread.stop();
        algThread=null;
      }
    WCanvas.paint();
  }
  
  
  int stepNo=0;
  /** Button Step performs one step (on generation) of genetic algorithm */
  protected void actionStep()
  {
    WCanvas.setPosition(stepNo,0);
    if(++stepNo==WalkParams.POSITIONS-1) 
      {
        population.newGeneration(0,5);
        stepNo=0;
        myCreature.joint.resetVar();  
      }
  }
  
  /** Defines run of genetic algorithm thread */
  public void run()
  {
    WCanvas.setCreature(myCreature);
    WCanvas.paint();
    try { Thread.sleep(STEP_PAUSE);
        } catch (InterruptedException e) {}
    while(true)
      {
         actionStep();   
         try { Thread.sleep(STEP_PAUSE);
             } catch (InterruptedException e) {}
      }
  }

  /** Finishing of the applet and destroying thread. */
  public void destroy()
  {
    if(algThread!=null)
      {
        algThread.stop();
        algThread=null;
      }
    super.destroy();
  }

  /** Finishing of the applet and destroying thread.
    * @exception Throwable successor can throw an exception
    */
  protected void finalize() throws Throwable
  {
    if(algThread!=null)
      {
        algThread.stop();
        algThread=null;
      }
    super.finalize();
  }

  /** Returns information about parameters of applet. */
  public String[][] getParameterInfo() 
  {
    String[][] info = 
     {
      {"BGCOLOR", "color (6 hexa digits)", "Background color, for example \"00ff40\"."},
     };
        return info;
  }

  /** Returns information about applet. */
  public String getAppletInfo() 
  {
    return "Title: Genetic algorithm developing movement \n"+
           "Author: Marek Obitko, obitko@email.cz \n"+
           "An applet shows GA evolving movement (walking) of artificial creature. Not finished yet. ";
  }


  /** For running of applet without browser or appletviewer (it is then possible
    * to run this applet as a class directly).
    * @see GAApplication
    */
  public static void main(String[] args)
  {
    Applet applet = new WalkGA();
    Frame frame = new GAApplication("Genetic Algorithm Developing Movement", applet, 540, 430);
  }
}